/*	
 *	AntilagBullet overrides an ROBullet to pass hits through an AntilagRelay
 *	to the server, to the firing weapon, which then performs the hits instead
 *
 *	Server functions that require bullet data but are not a pawn hit are dealt
 *	with by passing the server a "shot segment", which are the end points
 *	of a bullet's flight path.  The server will then replicate impact fx,
 *	deal suppression & bullet cracks and attempt to hit any players outside
 *	the firing client's relevancy using those end points.  If the bullet 
 *	flies a certain period of time without impacting anything, it breaks off
 *	a segment at its mid-air location, so that suppression effects don't have 
 *	to wait until the end of its flight (which may be quite some time later)
 *
 *		-Mekhazzio 
 */		

class AL_`{ALBULLETNAME} extends `{ALBULLETNAME};

var	float			PenDepthRemaining, MaxPenDepth;
var	array<ImpactInfo>	PreviousImpactList;
var 	bool			bAdvancedBallistics;
var 	int			LastSentEndIndex;
var	AntilagRelay		ALRelay;

//Notify the server, via the relay, that this bullet has just hit Target
function PassHitToRelay(Actor Target, vector HitLocation, TraceHitInfo HitInfo, optional bool bSpawnOldProjectile = false)
{
	local float AdjustedDamage;
	local byte ShotPower;

	`ifdef(ALTRACER)
		return; //tracers don't do any effects
	`endif
	
	if (bAdvancedBallistics) {
		AdjustedDamage =  PenDepthRemaining / MaxPenDepth;
		if ( default.Damage < 100) //powerful rounds don't lose damage to speed
			AdjustedDamage = FMin(VSizeSq(Velocity) / (default.Speed * default.Speed) , AdjustedDamage);
		ShotPower = AdjustedDamage * 255; 
	} else
		ShotPower = 255;
		
	if (bSpawnOldProjectile)
		ShotPower = Min(255, (VSize(Velocity) * 255)  / default.Speed);

	//`ALlog("PassHitToRelay" @ ALRelay @ Instigator @ Owner @ "->" @ Target @ "@" @ HitLocation @ HitInfo.BoneName @ ShotPower)              	

	if (ALRelay != none)
	{
		ALRelay.DoALHit(Target, HitLocation, HitInfo,bSpawnOldProjectile,ShotPower);
	}
	else
	{
		`ALlog("ERROR: PassHit couldn't find a relay for" @ self @ owner)       	
	}
}

//sends the end points of a shot segment, for impact effects, suppression, etc
function PassSegmentToRelay()
{
	local int StartIdx, EndIdx, i;
	local Actor HitActor1, HitActor2;
	
	`ifdef(ALTRACER)
		return; //tracers don't do any effects
	`endif
	
	EndIdx = PreviousImpactList.Length - 1;
	ResetSegmentTimer();

	for(i = EndIdx - 1; i >= 0; i--)
	{
		if ( ROPawn(PreviousImpactList[i].HitActor) != none )
		{
			if (HitActor1 == none)
				HitActor1 = PreviousImpactList[i].HitActor;
			else if (HitActor2 == none)
				HitActor2 = PreviousImpactList[i].HitActor;
		}
		else
		{
			StartIdx = i;
			break;
		}
	}
	
	if (VSize( PreviousImpactList[EndIdx].HitLocation - PreviousImpactList[StartIdx].HitLocation) < 3.0)
		return;  //don't bother sending tiny blips and/or a stopped bullet

	//`ALlog("PassSegmentToRelay" @ ALRelay @ Instigator @ Owner @ self)              	

	if (ALRelay != none)
	{
		ALRelay.DoALSegment(PreviousImpactList[StartIdx], PreviousImpactList[EndIdx], HitActor1, HitActor2);
	}
	else
	{
		`ALlog("ERROR: PassSegment couldn't find a relay for" @ self @ owner)              	
	}
}

//has to be overridden to replace TakeDamage with PassHitToRelay
simulated function ProcessBulletTouch(Actor Other, Vector HitLocation, Vector HitNormal, PrimitiveComponent OtherComp)
{
    	local ImpactInfo 	CurrentImpact;
   	local array<ImpactInfo> HitInfos;
    	local vector 		StartTrace, EndTrace;
   	local int 		i;

	`ifdef(ANTILAGGFX)
		local color debugcolor;
		debugcolor.R = 255; debugcolor.A = 255;
	`endif
	
	//`ALlog(self @ "ProcessBulletTouch:" @ Other @ "(" @ Other.class @ ")" )

	`ifdef(ALTRACER)
		return; //tracers are graphics only
	`endif

       	if( ROPawn(Other) == none && ROAerialReconPlane(Other) == none)
        {
        	return;
        }	
 
       	//Since these bullets no longer automatically fizzle when they hit something,
       	//it's a good idea to double-check that we don't hit the same target twice
       	for (i=0; i < PreviousImpactList.Length; i++)
       	{
       		if (Other == PreviousImpactList[i].HitActor)
       		{
			//`ALlog("ALProcessBulletTouch aborting, already hit" @ Other)
       			return;	
       		}		
       	}
 
       	CurrentImpact.HitActor		= Other;
       	CurrentImpact.HitLocation	= HitLocation;
       	CurrentImpact.HitNormal		= HitNormal;
       	CurrentImpact.RayDir		= Normal(Velocity);
       	CurrentImpact.HitInfo.HitComponent	= OtherComp;
	
          	// Convert Trace Information to ImpactInfo type.

       	StartTrace = HitLocation;
       	if (Other.GetMyPawn() != none)
       	{
		EndTrace = StartTrace +  CurrentImpact.RayDir * (Other.GetMyPawn().CylinderComponent.CollisionRadius * 6.0);
	}
	else
	{
		EndTrace = StartTrace + CurrentImpact.RayDir * 1500.0;
	}

	CurrentImpact.StartTrace = StartTrace;

    	GetAllHitBones(CurrentImpact, HitInfos, EndTrace, StartTrace, vect(0,0,0));

        if( HitInfos.Length == 0 )
       	{
        	return;  //the precision trace returned nothing
        } 

       	for (i = 0; i < HitInfos.Length; i++)
       	{
		if (HitInfos[i].HitActor != Other) //we'll hit this guy later when we Touch him
		{
			//`ALlog("ProcessBulletTouch:" @ HitInfos[i].HitActor @ " is not " @ Other $", hope I hit him later")
			continue;  //it appears that GetAllHitBones won't actually return hits on another actor, but just in case
		}	

              	HitInfos[i].RayDir = CurrentImpact.RayDir;
   		if( ROPawn(Other) != none)
   		{
              		// TODO: Figure out why the code isn't returning the right physical materials for ROPawns
                  	// Instead of just setting the value here - Ramm
                   	// at least it's a hack with honesty - Mekh
                   	HitInfos[i].HitInfo.PhysMaterial = PhysicalMaterial'ENV_Surface_Types.Physical_Materials.Flesh_PM';
               	}
`ifdef(ANTILAGGFX)
		`ALSphere(HitInfos[i].HitLocation,1,12,255,0,0)
		`ALLine(HitInfos[i].HitLocation + CurrentImpact.RayDir*8,HitInfos[i].HitLocation - CurrentImpact.RayDir*8, 255, 0,0)
		StartTrace.X = 0;
		StartTrace.Y = 0;
		StartTrace.Z = 1 - i;
		`ALString(HitInfos[i].HitLocation - (CurrentImpact.RayDir cross vect(0,0,1))*2 - StartTrace, HitInfos[i].HitInfo.BoneName, none, debugcolor)
`endif          
      	
		PassHitToRelay(Other, HitInfos[i].HitLocation, HitInfos[i].HitInfo);
       	}
       
       	PreviousImpactList.AddItem(CurrentImpact);
}

//the stock penetration code is tangled and looks pretty slow
//It's less work to just replace it rather than try to interface with it
//overview: on wall touch, find the far side, move the bullet there, let it continue on with normal flight
//advanced ballistics is included just because, once I'm overriding this function already, it's trivial to add
simulated singular event HitWall(vector HitNormal, actor Wall, PrimitiveComponent WallComp, optional PhysicalMaterial WallPhysMaterial)
{
	local float 		MaxPenDistance;
	local vector		ShotDir;
	local vector 		InverseTraceStart, InverseTraceEnd;
	local vector 		TraceLoc, TraceNorm;
	local Actor 		TraceActor, InteriorActor;
	local TraceHitInfo	THInfo;
	local ECollisionType	OldColType;
	local vector 		InteriorLoc, InteriorNorm;
	local ImpactInfo	ExitImpact;
	local vector 		NewProjectileLocation;
	local float		ModifiedPenDepth, MaterialPenetrationModifier;
	local int 		i;
	local float		HitAngle, MaxSkipAngle, SkewRatio;
	local vector		ReflectDir, OrthoDir;
	
`ifdef(ANTILAGGFX)
	local byte debugcolor;
	
	debugcolor = 255 * PenDepthRemaining / MaxPenDepth;
`endif

	MaterialPenetrationModifier = GetMatPenMod(WallPhysMaterial);
    	ShotDir = Normal(Velocity);

`ifndef(ALTRACER)
	//add this impact (the entry point) to the bullet's list of hits
	i = PreviousImpactList.Length;
	PreviousImpactList.Add(1);  
	PreviousImpactList[i].HitActor 		= Wall;
	PreviousImpactList[i].HitLocation 	= Location;
	PreviousImpactList[i].HitNormal 	= HitNormal;
	PreviousImpactList[i].RayDir 		= Normal(Velocity);       	
	PreviousImpactList[i].HitInfo.PhysMaterial = WallPhysMaterial;
	PreviousImpactList[i].HitInfo.HitComponent = WallComp;
	

	
	PassSegmentToRelay(); //send off the previous segment
	//`ALlog(self @ "HitWall" @ Wall @ "@" @ ShotDir dot HitNormal @ "angle " @ WallPhysMaterial @ MaterialPenetrationModifier $ "x with " @ PenDepthRemaining @ "remaining")
`else
	`ALSphere(Location,6,6, 255,0,255)       	
`endif

	if (!Wall.bStatic)  //bullet has hit damagable geometry
	{
            	if( Wall.GetMyVehicle() != none)
            	{
			`ifdef(ALATBULLET)
            			PassHitToRelay(Wall, Location - Wall.Location, PreviousImpactList[i].HitInfo, true);
            			//`ALlog("AT rifle spawning old projectile at" @ Location)
            			DeflectFX(Location, Normal(Velocity) * -1);
            			return;  
			`endif
			`ifndef(ALTRACER) //tracers shouldn't try to hurt tanks
	            		if (BulletHitTank(i, ROVehicleTank( Wall.GetMyVehicle() )  )  )
	            		{
	            			Destroy();
	            			return; //bullet stopped inside the tank
	            		}
	            	`endif
        	} 
        	else //destructible objects need to be hit
        		PassHitToRelay(Wall, Location, PreviousImpactList[i].HitInfo);
        }


     	//this whole section is full of arbitrary numbers chosen for general feel
     	if (bAdvancedBallistics)
     	{
		bBlockedByInstigator = true; //now that we've hit something, the shooter is fair game!
     		
     		Speed = VSize(Velocity);
		ReflectDir = Normal(-2 * (Velocity dot HitNormal)*HitNormal + Velocity); 
		OrthoDir = Normal(Velocity cross HitNormal);
		
	    	`ALLine(Location,Location-ShotDir*25, 255 - debugcolor, debugcolor, 0)
	    	
	    	HitAngle = ShotDir dot HitNormal;
	    	MaxSkipAngle = -1 + FMin(0.9, MaterialPenetrationModifier); //ricochets are more likely on harder surfaces
	    	if (HitAngle < 0 && HitAngle > MaxSkipAngle * FRand()) //ricochets are more likely at steeper angles
	    	{
			`ALBox(Location, vect(1,1,1), 255, 255, 255, true)
			
		    	SkewRatio = FMin(Abs(HitAngle)*1.5, 1.0); //steeper angles result in more randomness
		    	ShotDir = ReflectDir  +  OrthoDir*RandRange(-SkewRatio, SkewRatio)  +  ShotDir*RandRange(-SkewRatio, SkewRatio);
		    	ShotDir = Normal(ShotDir);
		    	
			Speed *= FMin(0.8, (1 - SkewRatio)); //never lose more than 80% speed
			PenDepthRemaining -= MaxPenDepth * FMax(0.15, Abs(HitAngle));  //ricochets shed lots of power
`ifdef(ANTILAGGFX)
			debugcolor = 255 * PenDepthRemaining / MaxPenDepth;
	    	 	`ALLine(Location,Location+ShotDir*25, 255 - debugcolor, debugcolor, 0)
`endif			
			PreviousImpactList[i].HitActor = none;  //don't make another bullethole on this location
			Velocity = ShotDir * Speed;
			
			//`ALlog("Ricochet:"@ Speed $"uu/s @" @ HitAngle @ PenDepthRemaining $"uu pen w/ Skew:"@SkewRatio)
			
			`ifdef(ALTRACER) //tracers don't use pendepth
				if (Speed < 0.1 * default.Speed)
					Explode(Location, HitNormal);  //low power tracers look bizarre without a proper tracer fx in the game
				SetRotation(rotator(Velocity)); //point tracer in the right direction
			`else
				if (PenDepthRemaining < 0.01) //clean up the bullet if it's lost all its power on the ricochet
				{
					`ALBox(Location, vect(2,2,2), 255, 0, 0, true)
					Destroy();
					return;				
				}
			`endif
									
			return;
	    	}
	    	//not bouncing, so refract through the material we're hitting
	    	SkewRatio = (1 - Abs(HitAngle)) * 0.4;
	    	ShotDir += OrthoDir * RandRange(-SkewRatio, SkewRatio);
	    	ShotDir -= ReflectDir * SkewRatio*2*FRand();
		ShotDir = Normal(ShotDir);
		
		SkewRatio = 0.6 * (1 - (ShotDir dot Normal(Velocity)));  //subtract power based on how far it actually skewed
		PenDepthRemaining -= MaxPenDepth * SkewRatio;
		Speed -= MaxSpeed * SkewRatio;
		//`ALlog("Refract:"@ Speed $"uu/s @" @ HitAngle @ PenDepthRemaining $"uu pen w/ Skew:"@SkewRatio)
	} //done with adv. ballistics

	`ifdef(ALTRACER)
		Explode(Location, HitNormal); //tracers don't penetrate
	`endif

	//penetration check begins
    	MaxPenDistance = PenDepthRemaining * MaterialPenetrationModifier;
 	InverseTraceEnd = Location;
	InverseTraceStart = Location + ShotDir * MaxPenDistance;
     	`ALStar(InverseTraceStart, 5, 255 - debugcolor, debugcolor, 0)
	`ALLine(InverseTraceStart,InverseTraceEnd, 255 - debugcolor, debugcolor, 0)
	
	for (i=0; i < `{ALMAXPENTESTS}; i++ )
	{
        	InverseTraceStart = Location + ShotDir * MaxPenDistance/`{ALMAXPENTESTS} * (`{ALMAXPENTESTS} - i);

		TraceActor = Trace(TraceLoc, TraceNorm, InverseTraceEnd, InverseTraceStart, true, vect(0,0,0), THInfo, TRACEFLAG_Bullet);
		//`ALlog("HitWall Trace:" @ InverseTraceStart @ "to" @ InverseTraceEnd @"hits" @ TraceActor  @ TraceLoc @ "="@ VSize(TraceLoc-InverseTraceStart)$"uu" @ Normal(Velocity) dot TraceNorm)

 		if (TraceActor == none)
		{
			break;	//we've traced back to the entry point, so we're done		
		}   
		if (TraceLoc ==	InverseTraceStart)
		{
			continue; //this trace went nowhere (generally started inside something) so skip
		}
		if (ROPawn(TraceActor) != none) 
		{
			continue; //pawns are never a wall exit point
		}
		if (ShotDir dot TraceNorm <= 0)
		{
			continue;  //A reverse-facing polygon isn't an exit point
		}
		
		//if we're inside an overlapping object, then it isn't an exit point
    		OldColType = Wall.CollisionType;
		Wall.SetCollisionType(COLLIDE_BlockAllButWeapons);  //turn off trace collision on the original object hit
		InteriorActor = Trace(InteriorLoc, InteriorNorm, Location, TraceLoc, true, vect(0,0,0),, TRACEFLAG_Bullet);
		Wall.SetCollisionType(OldColType);  //collision back on

		if (InteriorActor != none && ShotDir dot InteriorNorm <= 0)
		{
			`ALBox(Location, vect(2,2,2), 50, 50, 255, true)
			continue;  //A reverse-facing polygon here means we're inside an overlapping object, so not an exit point
		}
		
		`ALCylinder(InverseTraceStart, TraceLoc, 1 + i, 6, 255, 255 , 255, true)		
		ExitImpact.HitActor 	= TraceActor;
		ExitImpact.HitLocation 	= TraceLoc;
		ExitImpact.HitNormal 	= TraceNorm;
		ExitImpact.StartTrace	= InverseTraceStart;
		ExitImpact.RayDir 	= ShotDir;
		ExitImpact.HitInfo	= THInfo;

		NewProjectileLocation = TraceLoc;
	}
 	
 	if (ExitImpact.HitActor == none)  //no exit point means the bullet stops here
 	{
		`ALBox(Location, vect(4,4,4), 0, 255, 255, true)
		Destroy();
		return;					 		
	}

	//this fixes the issue where non-matching surface materials penetrate differently from opposite sides
	MaterialPenetrationModifier = FMin(MaterialPenetrationModifier, GetMatPenMod(ExitImpact.HitInfo.PhysMaterial));

	ModifiedPenDepth = VSize(NewProjectileLocation - Location) / MaterialPenetrationModifier;
	PenDepthRemaining -= ModifiedPenDepth;
//	`ALlog("HitWall went through" @ VSize(NewProjectileLocation - Location) $ "uu at" @ MaterialPenetrationModifier $ "x used" @ ModifiedPenDepth @ "with" @ PenDepthRemaining @ "remaining")
	
	if (PenDepthRemaining < 0.01)
	{
//		`ALlog("HitWall: bullet has run out of penetration power, shutting down")
		`ALBox(Location, vect(4,4,4), 255, 0, 0, true)
		Destroy();
		return;				
	}

	`ALSphere(Location,1 + 5 * debugcolor / 255,8, 255 - debugcolor, debugcolor, 0)
	`ALLine(Location,Location+ShotDir*80, 0, 255, 255)

	if (bAdvancedBallistics)
	{
		Velocity = ShotDir*Speed;
	}
	
       	ExitImpact.bExitImpact = true;
	PreviousImpactList.AddItem(ExitImpact);

	//Move the bullet to the exit point to let it continue on its path
	SetLocation(NewProjectileLocation);

`ifdef(ANTILAGGFX)
	debugcolor = 255 * PenDepthRemaining / MaxPenDepth;
	`ALSphere(Location,1 + 5 * debugcolor / 255,12, 255 - debugcolor, debugcolor, 0)       	
`endif	
}

//split out from ROBullet.HitWall & cleaned up
//returns true if the bullet should stop at the tank
function bool BulletHitTank(int ImpactListIndex, ROVehicleTank ROVT)
{
	local ImpactInfo	CurrentImpact;
	local vector		StartTrace, EndTrace;
	local array<ImpactInfo>	HitInfos;
	local int		i, HitIndex, ArmorZoneIndex;
	local bool		bIgnoreNextArmor, bStopBullet;

	if(  ROVT == none || DamageRadius != 0 || ImpactListIndex >= PreviousImpactList.Length)
		return false;
	
	CurrentImpact = PreviousImpactList[ImpactListIndex];
	
	StartTrace = Location;
	EndTrace = StartTrace + CurrentImpact.RayDir * (ROVT.CylinderComponent.CollisionRadius * 6.0);
	`ALLine(StartTrace,EndTrace, 255, 0, 0)
	
	GetAllHitBones(CurrentImpact, HitInfos, EndTrace, StartTrace, vect(0,0,0));
	
	if (HitInfos.Length < 1)
		return false; 	//no bones hit
	
	bIgnoreNextArmor = false;
	bStopBullet = false;
	for (i = 0; i < HitInfos.Length; i++)
	{
		if( HitInfos[i].HitInfo.BoneName == '' )
		{
			continue;
		}
		HitIndex = ROVT.VehHitZones.Find('ZoneName',HitInfos[i].HitInfo.BoneName);
		ArmorZoneIndex = ROVT.ArmorHitZones.Find('ZoneName',HitInfos[i].HitInfo.BoneName);
		
		// Handle the shot hitting a person in the vehicle
		if( HitIndex >= 0 && (ROVT.VehHitZones[HitIndex].VehicleHitZoneType == VHT_CrewBody ||
			ROVT.VehHitZones[HitIndex].VehicleHitZoneType == VHT_CrewHead) )
		{
			`ALSphere(HitInfos[i].HitLocation,4,12,255,0,0)
			HitInfos[i].HitInfo.PhysMaterial = PhysicalMaterial'ENV_Surface_Types.Physical_Materials.Flesh_PM';
			PassHitToRelay(ROVT, HitInfos[i].HitLocation, HitInfos[i].HitInfo);
			bStopBullet = true; //don't ricochet off someone's head
		}
			// The shot hit an instant penetration zone, keep looping and see if it would hit a human
		else if ( ArmorZoneIndex >= 0 && ROVT.ArmorHitZones[ArmorZoneIndex].bInstaPenetrateZone )
		{
			bIgnoreNextArmor = true;
			bStopBullet = true; //don't ricochet after going inside the tank
		}
		else if( ArmorZoneIndex >= 0 && bIgnoreNextArmor )
		{
			bIgnoreNextArmor = false;
		}
		else
			break;
	}
	return bStopBullet;
}

//overridden to announce endpoints for the suppression trace at the end of bullet life
simulated event Destroyed()
{
	DoMidAirSegment();
	Super.Destroyed();
}

//returns the material penetration modifier for a material (cut & pasted from ROProjectileWeapon.RecursePenetration)
function float GetMatPenMod(PhysicalMaterial WallPhysMaterial)
{
	local ROPhysicalMaterialProperty 	PhysicalProperty;
	local float 				MaterialPenetrationModifier;

	if (WallPhysMaterial != none)
	{
		PhysicalProperty = ROPhysicalMaterialProperty(WallPhysMaterial.GetPhysicalMaterialProperty(class'ROPhysicalMaterialProperty'));
	}
	if( PhysicalProperty != none )
	{
	   	MaterialPenetrationModifier = class'ROImpactEffectComponent'.static.GetMaterialPenetrationModifier(PhysicalProperty.MaterialType);
	}
	else
	{
	   	MaterialPenetrationModifier = class'ROImpactEffectComponent'.static.GetMaterialPenetrationModifier(EMT_Default);
	}
	
	return MaterialPenetrationModifier;
}

//overridden to set up pen depth & impact list
simulated function Init(vector Direction)
{
	Super.Init(Direction);	

	if (ROWeapon(Owner) != none) 
	{
		MaxPenDepth = ROWeapon(Owner).default.PenetrationDepth;
		PenDepthRemaining = MaxPenDepth;
	}

	ResetSegmentTimer();
	AddNullImpactToList(); 	//start point
	PreviousImpactList[0].HitActor = Owner; //hitactor == firingweapon denotes the initial segment
}

//adds an impact with no actor to the list - for midair segments
function AddNullImpactToList() 
{
	local int i;
	
	i = PreviousImpactList.Length;
	PreviousImpactList.Add(1); 
	PreviousImpactList[i].HitLocation 	= Location;
	PreviousImpactList[i].RayDir 		= Normal(Velocity);       	
}

//if the bullet hasn't hit anything after a certain amount of time, it just
//goes ahead and sends out a segment, to prevent adding excessive delay
//to suppression & such
function DoMidAirSegment()
{
	//`ALlog("Bullet: MidAirSegment()" @ Location  @ "@" @ WorldInfo.TimeSeconds - CreationTime $ "s" )
	AddNullImpactToList();
	PassSegmentToRelay();
}

function ResetSegmentTimer()
{
	SetTimer(`{ALBULLETSEGMENTTIME} , false, 'DoMidAirSegment' );
}

//since it's never used for anything else for Bullets, I can override this to pass the relay actor in
simulated function ProcessTouch(Actor Other, Vector HitLocation, Vector HitNormal)
{
	if (AntilagRelay(Other) != none)
	{
		ALRelay = AntilagRelay(Other);
		bAdvancedBallistics = ALRelay.bAdvancedBallistics;	
	}	
}

defaultproperties
{
	RemoteRole=ROLE_None
`ifdef(ALTRACER)
	bAdvancedBallistics=true  //saves me from having to override functions outside the scope of Antilag - I assume most servers will have it on, anyway
`endif
`ifdef(ALDMGTYPEHACK)
	MyDamageType=class'`{ALDMGTYPEHACK}'
`endif
}