/*	
 *	AntilagWeapon overrides an ROProjectileWeapon to fire locally authoritative
 *	bullets that then replicate hit messages through an AntilagRelay to the server,
 *	and back to this weapon, which then performs the hits directly
 *
 *		-Mekhazzio 
 */		

`ifdef(ALCLASSOVERRIDE)
class AL_`{ALCLASSOVERRIDE} extends ROWeap_`{ALWEAPONNAME};
`else
class AL_`{ALWEAPONNAME} extends ROWeap_`{ALWEAPONNAME};
`endif
 
var 	float 		MIMaxShootingRange; 	//clients can't see this from MapInfo, so it has to replicate
var 	Antilag 	ALMutator;		//reference back to the mutator, to do server-side global stuff
var 	bool		bAdvancedBallistics;	//perform ricochets & such
var 	class<ROProjectile>	OriginalRO2Projectile;	//for AT weapons
var 	Controller	InstigatorController;  	//controller that's doing the firing, needed for dealing post-death damage
var 	vector		LastShotSourceLoc;	//for post-death effects
var	AntilagRelay	ALRelay;		//for communicating hits back to the server

replication
{
	if ( bNetDirty && bNetOwner && Role==ROLE_Authority )
		MIMaxShootingRange, bAdvancedBallistics, ALRelay;
}

//finds the controller's relay and (server only) points the relay's delegates to this weapon
simulated function UpdateRelayLinks()
{
	local AntilagRelay relay;

	//find the shooter's relay if the old one is wrong
	if (ALRelay == none || (Instigator != none && ALRelay.Controller != Instigator.Controller)  )
	{
		ALRelay = none;
		foreach DynamicActors(class'AntilagRelay', relay)
		{
			if (relay.Controller == Instigator.Controller)
				ALRelay = relay;
		}
		
		if (ALRelay == none)
		{
			`ALlog("PANIC: " @ InstigatorController @ Owner @ self @ "still has no relay when attacking!!")
			return;
		}
	}

	if (Role == ROLE_Authority)
	{	
		ALRelay.HitFunction = ServerALHit;
		ALRelay.SegmentFunction = ServerShotSegment;
	}
}

//keeps the relay linked to the proper weapon to avoid a race condition on bad connections
simulated state Bandaging
{
	simulated function WeaponEquipped()
	{
		UpdateRelayLinks();
		Super.WeaponEquipped();
	}
}
simulated state WeaponEquipping
{
	simulated function WeaponEquipped()
	{
		UpdateRelayLinks();
		Super.WeaponEquipped();
	}
}
simulated function WeaponEquipped()
{
	UpdateRelayLinks();
	Super.WeaponEquipped();
}

//server-side
//responds to a damage notification
function ServerALHit(Actor Target, vector HitLocation, TraceHitInfo HitInfo, optional bool bIsMeleeAttack = false, optional byte ShotPower = 0)
{
	local class<ROProjectile>	ProjectileClass;
	local class<DamageType> 	DamageType;
	local vector			Momentum, FireDir;
	local float 			Damage, ChargePct;

	ProjectileClass = class<ROProjectile>(GetProjectileClass());

	if (Target == none || InstigatorController == none || (!bIsMeleeAttack && ProjectileClass == none))
	{
		//`ALlog("NOTICE: Invalidated shot from" @ self @ Instigator @ InstigatorController @"at" @ Target @ "with" @ProjectileClass@"for referenced none")
		return;  //typically occurs because the target is already dead
	}

	if (ROPawn(Target) != none && VSize(Target.Location - HitLocation) > `ALINVALIDATEDISTANCE )
	{
		//`ALlog("NOTICE: Invalidated shot at" @ Target @ "due to gap size")
		return; //stops laggy shots from hitting respawned players
	}

	if ( Instigator == none && !IsInsideGracePeriod(VSize(HitLocation - LastShotSourceLoc))  )
	{
		//`ALlog("NOTICE: Invalidated shot at" @ Target @ "due to a dead shooter")
		return;  //keeps a reaction advantage for low pings
	}

	FireDir = Normal(HitLocation - LastShotSourceLoc); 
	
	if (bIsMeleeAttack && ShotPower > 1) //I really don't want to touch the vehicle penetration code, so I just spawn a stock ATR bullet against the tank
	{					//Carries the speed over to make sure it's not full power at all ranges
		SpawnOldRO2Projectile(Target, HitLocation, FireDir, ShotPower);
		return;
	}
	if (bIsMeleeAttack)
	{
		Momentum = FireDir;
		DamageType = InstantHitDamageTypes[MELEE_ATTACK_FIREMODE];
		ChargePct = FClamp((WorldInfo.TimeSeconds - MeleeHoldStartTime) / MeleeChargeTime, 0.0, 1.0);
		Damage = Lerp(MeleeAttackDamage, MeleeAttackChargeDamage, ChargePct);
		// Add speed based damage
		Damage += MeleeMovingDamageMod * MeleeAttackDamage;
		// Apply small random adjustment
		Damage = RandRange(Damage - MeleeAttackRandomDamage, Damage + MeleeAttackRandomDamage);

		MeleeFFNotify(ROPawn(Target).Controller);
		WeaponPlaySound(MeleeAttackHitFleshSound);
	}
	else  //is boolet
	{
		Momentum = ProjectileClass.default.MomentumTransfer * FireDir;
		DamageType = ProjectileClass.default.MyDamageType;
		Damage = ProjectileClass.default.Damage;
		if (bAdvancedBallistics && ShotPower > 0)
			Damage = Lerp(Damage/4, Damage, float(ShotPower) / 255); // 3/4ths of bullet damage scales with power remaining

		//spawn hit effects for bullets
		//SetFlashLocation(HitLocation);
		SetReplicatedImpact(HitLocation, FireDir, LastShotSourceLoc, class, 0.0, false );
	}

	//`ALlog("ServerALHit:" @ InstigatorController @ Instigator @ self @ "->" @ Target @ DamageType @ Damage @ HitInfo.BoneName )
	
	Target.TakeDamage(Damage, InstigatorController, HitLocation, Momentum, DamageType, HitInfo, self);
}

//server-side
//replicates bulletholes and performs suppression & bullet whips between two points
//also does a bullet trace to hit anyone between the two points, unless they're passed as Ignore1 or Ignore2
//this allows for overpenetration hits targets the shooter doesn't have relevancy on
//a bit of a hack, but lighter on bandwidth than a robust solution, and it's rare to even hit a 2nd guy, much less a 3rd
function ServerShotSegment(ImpactInfo Start, ImpactInfo End, optional Actor Ignore1, optional Actor Ignore2)
{ 
	local vector 		FireDir, TraceLoc, TraceNorm, StartTrace, EndTrace;
	local ROPawn		ROP;
	local Actor		TraceActor;
	local TraceHitInfo	THInfo;
	local ImpactInfo	CurrentImpact;
	local array<ImpactInfo> HitInfos;
	local vector 		WhizDistOffset;
	local int		i;
	
	//`ALlog(self @ "SSSeg:"@ Start.HitActor @ End.HitActor @ "=" @ VSize(End.HitLocation - Start.HitLocation) $"uu" @ "from" @ Instigator @ "("@Ignore1$","@Ignore2@")" )

	if ( InstigatorController.Pawn == none  && !IsInsideGracePeriod(VSize(End.HitLocation - LastShotSourceLoc)) )
	{
		//`ALlog("NOTICE: Invalidated segment due to a dead shooter")
		return;  //keeps the reaction advantage for low pings
	}
		
	FireDir = End.HitLocation - Start.HitLocation;

	if ( Start.HitActor != none && Start.HitActor != self) {  //self denotes the shot origin
		//SetFlashLocation(Start.HitLocation);			//segment start, if it exists, is an exit impact
		SetReplicatedImpact(Start.HitLocation, FireDir * -1, Instigator.Location, class, 0, false );
	}
	if ( End.HitActor != none) {
		//SetFlashLocation(End.HitLocation);
		SetReplicatedImpact(End.HitLocation, FireDir, Instigator.Location, class, 0, false );
	}
	
	StartTrace = Start.HitLocation;
	EndTrace = End.HitLocation;
	
	if (Start.HitActor != self) //actor=self means the first segment, so no relevancy backstop - everything in LOS should already be relevant
	{
		foreach TraceActors(class'ROPawn', ROP, TraceLoc, TraceNorm, EndTrace, StartTrace,,THInfo, TRACEFLAG_Bullet)
		{	
			TraceActor = ROP;
			if (TraceActor != Ignore1 && TraceActor != Ignore2)
			{	
			        CurrentImpact.HitActor		= TraceActor;
			        CurrentImpact.HitLocation	= TraceLoc;
			        CurrentImpact.RayDir		= FireDir;
			        CurrentImpact.HitNormal		= TraceNorm;
			        CurrentImpact.HitInfo		= THInfo;
				
		     	   	GetAllHitBones(CurrentImpact, HitInfos, End.HitLocation, StartTrace, vect(0,0,0));
	
			       	for (i = 0; i < HitInfos.Length; i++)
		        	{
		        		if ( HitInfos[i].HitActor == Ignore1 || HitInfos[i].HitActor == Ignore2 || HitInfos[i].HitActor != TraceActor)
		        			continue;
					//`ALlog("ServerShotSegment: Relevancy backstop hit" @ Instigator @ "on:" @ TraceActor @ HitInfos[i].HitInfo.BoneName )
		               		HitInfos[i].HitInfo.PhysMaterial = PhysicalMaterial'ENV_Surface_Types.Physical_Materials.Flesh_PM';
					ServerALHit(HitInfos[i].HitActor, HitInfos[i].HitLocation, HitInfos[i].HitInfo);
				}
			}
		}
	}
	
	if (VSize(End.HitLocation - Start.HitLocation) > Instigator.BulletWhizMaxDist)
		WhizDistOffset = FireDir * Instigator.BulletWhizMaxDist; //try not to overlap warnpawns on nearby segments
	
	WarnPawnsAlongFireLine(Start.HitLocation + WhizDistOffset, End.HitLocation);
}

`ifdef(ANTILAGGFX)
simulated exec function CheckAmmo()
{
	FlushPersistentDebugLines();
	Super.CheckAmmo();	
}
`endif

//I've yet to meet one that can outsmart boolet
simulated function Projectile ProjectileFire()
{
	local vector		RealStartLoc;
	local Projectile	SpawnedProjectile;
	local class<ROProjectile>	ProjectileClass;
	local vector 	 	BarrelDir;

`ifndef(ALHMG)
	local vector 		TraceLoc, TraceNorm, BarrelStartLoc;
`endif

	if (Instigator == none)
	{
		`ALlog("ERROR: ProjectileFire: this weapon ("$self$") has no shooter?!")
		return none;
	}

    	if( Instigator.IsFirstPerson() )
		HandleRecoil();

	// tell remote clients that we fired, to trigger effects
	IncrementFlashCount();

	InstigatorController = Instigator.Controller;
	UpdateRelayLinks();  //make sure the relay delegates link to this weapon
	LastShotSourceLoc = GetPhysicalFireStartLoc();

	if (Role == ROLE_Authority)
	{				
		if ( ROPlayerController(Instigator.Controller) != none && ROPlayerController(Instigator.Controller).StatsWrite != none )
		{ 
			ROPlayerController(Instigator.Controller).StatsWrite.AddToShots(1, InvIndex);
		}  	
	}		
		
	//spawn a projectile if this is the active client firing, or if it's AI controlled and we're the server
	//(post-DeMorgan's)
	if(!Instigator.IsLocallyControlled() && ( Role != ROLE_Authority || AIController(InstigatorController) == none ))
	{
		return none;
	}	
	
	ProjectileClass = class<ROProjectile>(GetProjectileClass());
	// get the gun position info

`ifndef(ALHMG)  //mounted weapons don't need to check start point
	BarrelStartLoc = GetBarrelBreechLocation(BarrelDir, RealStartLoc);

	if (!Instigator.bIronSights)  //hipfire needs to be checked to not spawn the bullet inside a wall
	{
	  	`ALStar(BarrelStartLoc, 2, 255,255,0)
		`ALLine(BarrelStartLoc, RealStartLoc, 255,255,0)
		`ALStar(RealStartLoc, 1, 255,255, 0)
		
	        if( Trace(TraceLoc, TraceNorm, RealStartLoc, BarrelStartLoc, true,,, TRACEFLAG_Bullet) != none )
		{
			RealStartLoc = TraceLoc;
			`ALSphere(BarrelStartLoc,2,12, 255, 255, 0)
		}
	}
	`ALStar(RealStartLoc, 2, 0,255, 255)
	`ALLine(RealStartLoc, RealStartLoc + BarrelDir*5, 0,255,255)
`else
	GetBarrelBreechLocation(BarrelDir, RealStartLoc);
`endif

	SpawnedProjectile = Spawn(ProjectileClass, self,, RealStartLoc);
	if (SpawnedProjectile == none)
	{
		`ALlog("ERROR: ProjectileFire spawning failed:" @self @ProjectileClass@RealStartLoc)
		return none;	
	}
	
	SpawnedProjectile.LifeSpan = MIMaxShootingRange / ProjectileClass.default.Speed;
	SpawnedProjectile.Init(BarrelDir);
	SpawnedProjectile.ProcessTouch(ALRelay, vect(0,0,0),vect(0,0,0)); //overriding this saves me a lot of pain with casting

	return SpawnedProjectile;
}

//overridden to use AL hit passing
//an amalgam of ROWeapon.MeleeAttackImpact and ROWeapon.GiveMeleeDamageTo
simulated function bool MeleeAttackImpact( optional int InDamage, optional out ROPawn OutVictimPawn, optional out TraceHitInfo OutHitInfo, optional bool bTest )
{
	local ROPawn			ROP;
	local vector			CheckDir, TraceLoc, EndTrace, StartTrace, TraceNorm;
	local Actor			TraceActor; 
	local TraceHitInfo		THInfo;
	local array<ImpactInfo>		HitInfos;
	local ImpactInfo		CurrentImpact;
	local int			i;
	local FracturedStaticMeshActor FracActor;
	
`ifdef(ANTILAGGFX)
	local color debugcolor;
	
	debugcolor.R = 255; debugcolor.A = 255;
`endif

	ROP = ROPawn(Owner);
	// Make sure Weapon/Pawn is still allowed to melee.  // Limit to 4 checks
	if( ROP == none || !ROP.CanDoMeleeImpact() || NumMeleeAttackAttempts > 4 )
	{
		return false;
	}

	InstigatorController = Instigator.Controller;
	UpdateRelayLinks();  //make sure the relay delegates link to this weapon
	LastShotSourceLoc = ROP.GetWeaponStartTraceLocation();

	//do attack only if this is the active client firing, or if it's AI controlled and we're the server
	//(post-DeMorgan's)
	if(!ROP.IsLocallyControlled() && ( Role != ROLE_Authority || AIController(ROP.Controller) == none ))
	{
		return false;
	}
	

	NumMeleeAttackAttempts++;

	//StartTrace = GetMeleeStartTraceLocation(); - this doesn't work for bots
	StartTrace = LastShotSourceLoc;
	CheckDir = vector(ROP.GetBaseAimRotation());
	EndTrace = StartTrace + CheckDir * MeleeAttackRange; //trace forward

  	`ALStar(StartTrace, 4, 255,255,0)
	`ALLine(StartTrace, EndTrace, 255,255,0)
	`ALStar(EndTrace, 2, 255,255,0)

	ROP = MeleeAttackFindPawnByFOV(CheckDir, THInfo, MeleeAttackRange); //run the melee autoaim
	TraceActor = Trace(TraceLoc, TraceNorm, EndTrace, StartTrace, true,,THInfo, TRACEFLAG_Bullet);

	//try the autoaim if the manual aim missed
	if (ROP != none &&  ROPawn(TraceActor) == none)
	{
		EndTrace = ROP.Mesh.GetBoneLocation(ROP.MeleeDamageBoneName); 
		TraceActor = Trace(TraceLoc, TraceNorm, EndTrace, StartTrace, true,,THInfo, TRACEFLAG_Bullet);
		`ALLine(StartTrace, EndTrace, 255,0,0)
		`ALStar(EndTrace, 2, 255,0,0)
	}

	if(TraceActor == none) //hit nothing
	{
		return false;
	}

	// Convert Trace Information to ImpactInfo type.
        CurrentImpact.HitActor		= TraceActor;
        CurrentImpact.HitLocation	= TraceLoc;
        CurrentImpact.RayDir		= Normal(EndTrace - StartTrace);
        CurrentImpact.HitNormal		= TraceNorm;
        CurrentImpact.HitInfo		= THInfo;
	
	if (ROPawn(TraceActor) != none)
	{
     	   	GetAllHitBones(CurrentImpact, HitInfos, EndTrace, StartTrace, vect(0,0,0));
 
         	if( HitInfos.Length == 0 )
        	{
			`ALSphere(EndTrace,4,8, 0, 225, 225)
                	return false;  //the precision trace returned nothing
            	} 

		WeaponPlaySound(MeleeAttackHitFleshSound);
	
	       	for (i = 0; i < HitInfos.Length; i++)
        	{
       			if (ROPawn(HitInfos[i].HitActor) == none) 
        		{
        			break;  //stop if we get to something that doesn't go squish
        		}
               		HitInfos[i].HitInfo.PhysMaterial = PhysicalMaterial'ENV_Surface_Types.Physical_Materials.Flesh_PM';
               		
			`ifdef(ANTILAGGFX)
				`ALSphere(HitInfos[i].HitLocation,1,12,255,0,0)
 				StartTrace.X = 0;	StartTrace.Y = 0;	StartTrace.Z = 1 - i;
 				`ALString(HitInfos[i].HitLocation - (CurrentImpact.RayDir cross vect(0,0,1))*2 - StartTrace, HitInfos[i].HitInfo.BoneName, none, debugcolor)
			`endif  
			              	
			ALRelay.DoALHit(TraceActor, HitInfos[i].HitLocation, HitInfos[i].HitInfo, true);
			
			if ( !bAllowMeleeToPenetrate )
			{
				break;
			}
         	}
		return true;	
	} 
	else if (!bHitWallThisAttack) //hit a wall
	{
		bHitWallThisAttack = true;
		// Sound
		WeaponPlaySound(MeleeImpactSound); // hit wall
		ALServerBashSound();
		// Particle
		class'RODmgType_MeleeBlunt'.static.SpawnHitEffect(Instigator, 0, vect(0,0,0), THInfo.BoneName, TraceLoc);

		// Fracture meshes if we hit them
		if(bAllowMeleeToFracture)
		{
			FracActor = FracturedStaticMeshActor(TraceActor);
			if(FracActor != None)
			{
				FracActor.BreakOffPartsInRadius(TraceLoc - (TraceNorm * 15.0), 35.0, 100.0, TRUE);
				//DrawDebugSphere(HitLoc - (WallImpact.HitNormal * 15.0), 35.0, 16.0, 255,0,0, TRUE);
			}
		}

		if (ROStaticMeshDestructible(TraceActor) != none)
			ALRelay.DoALHit(TraceActor, TraceLoc, THInfo, true);
	}

	MeleeFFNotify(Instigator.Controller);
	`ALSphere(TraceLoc,3,6,255,255,0)

	return false; 
}

unreliable server function ALServerBashSound() //get it to the server,  but don't care if it's dropped
{
	WeaponPlaySound(MeleeImpactSound);
}

//Force-feedback for an FPS?  Weird, but it's in the game, so I don't want to kill it...
simulated function MeleeFFNotify(Controller OtherCont)
{
     	local ROPlayerController PC;
     	PC = ROPlayerController(OtherCont);
	if( PC != none )
	{
		PC.ClientPlayForceFeedbackWaveform(class'ROWaveForms'.default.MeleeHit);
	}	
}

//overridden to initialize info
simulated event PostBeginPlay()
{
	local Antilag mut;
	local ROMapInfo ROMI;

	if (Role == ROLE_Authority)
	{	
		ROMI = ROMapInfo(WorldInfo.GetMapInfo());
		if (ROMI != none)
		{
			MIMaxShootingRange = ROMI.MaxShootingRange;
		}
		
		foreach DynamicActors(class'Antilag', mut)
		{
			ALMutator = mut;
			bAdvancedBallistics = mut.bAdvancedBallistics;
		}
		
`ifndef(ALHMG)  //mounted MGs don't have a controller (thus, no relay) when they spawn
		if (ALMutator != none && Instigator != none)
		{
				ALRelay = ALMutator.GetMyRelay(Instigator.Controller);
		}
`endif
	}
			
	Super.PostBeginPlay();	
}

simulated function InstantFire()
{
	ProjectileFire();
}

simulated function CustomFire()
{
	//skip the 50 meter hitscan thing & go straight to projectile firing
	//I assume it's there for server performance purposes
	//client-side handling renders that moot
	ProjectileFire();
}

simulated function vector GetBarrelBreechLocation(optional out vector BarrelDirection, optional out vector BarrelTipLocation)
{
	local vector RealStartLoc, BarrelStartLoc, BarrelDir;
	local ROPawn ROP;
	local float multiplier;
	
	RealStartLoc = GetPhysicalFireStartLoc();
	BarrelStartLoc = RealStartLoc;
	BarrelDir = vector(GetAdjustedAim(RealStartLoc));
	
	multiplier = 2.5;
	if ( Category == ROIC_Secondary ) 
		multiplier = 2.0; //pistols are shorter
		
	ROP = ROPawn(Owner);
	if (ROP != none)
	{
		BarrelStartLoc = RealStartLoc - Normal(BarrelDir) * ROP.CylinderComponent.CollisionRadius*multiplier;
	}
	
	BarrelDirection = BarrelDir;
	BarrelTipLocation = RealStartLoc;
	return BarrelStartLoc;	
}

//overridden to add the simulated flag now that clients are firing bullets
simulated function class<Projectile> GetProjectileClass()
{
	return (CurrentFireMode < WeaponProjectiles.length) ? WeaponProjectiles[CurrentFireMode] : None;
}

//server-side only
//prevents bullets already in the air from fizzling on shooter's death
function bool IsInsideGracePeriod(float ShotDistance)
{
	local float DiedTime, FlightTime;
	local class<Projectile> ProjClass;
	
	if (ALMutator == none || InstigatorController == none)
		return false;

	ProjClass = GetProjectileClass();
	if (ProjClass != none && ProjClass.default.Speed > 0)		
		FlightTime = ShotDistance / ProjClass.default.Speed;
	else
		FlightTime = `ALDEATHGRACEPERIOD ; 
		
	DiedTime = ALMutator.GetLastDiedTime(InstigatorController);
	
	//`ALlog("GracePeriod: " @ DiedTime @ "+" @ FlightTime @ "+" @ `ALDEATHGRACEPERIOD @ "  vs  " @ WorldInfo.TimeSeconds @ "("$ DiedTime + FlightTime + `ALDEATHGRACEPERIOD - WorldInfo.TimeSeconds $"s)" )
	
	//keep validating hits within the bullet's time of flight, plus a little extra wiggle room
	if (DiedTime + FlightTime + `ALDEATHGRACEPERIOD > WorldInfo.TimeSeconds)
		return true;
		
	return false;
}

//This is here just for AT rifles, so I don't have to override any vehicle penetration code
//Spawns an original RO2 projectile at an offset from a given target
//Some precision is lost in the vector replication, but only ~ 1/10th of a uu
function SpawnOldRO2Projectile(Actor Target, vector Offset, vector FireDir, byte ShotPower)
{
	local ROProjectile 	SpawnedProjectile;
	local float 		NewSpeed;
	local vector		SpawnPoint;
	
	SpawnPoint = Target.Location + Offset - FireDir*5;
	NewSpeed = OriginalRO2Projectile.default.Speed * ShotPower / 255; 
	//`ALlog("Spawned a" @ OriginalRO2Projectile @ (Target.Location + Offset) @ "with" @ ShotPower @ "=" @ NewSpeed);

	SpawnedProjectile = Spawn(OriginalRO2Projectile, self,, SpawnPoint);
	if (SpawnedProjectile != none)
	{
		SpawnedProjectile.Init(FireDir);
		SpawnedProjectile.Speed = NewSpeed;
		SpawnedProjectile.Velocity = FireDir*NewSpeed;
		if (ROAntiVehicleProjectile(SpawnedProjectile) != none)  //plink
			SpawnedProjectile.PlaySound(ROAntiVehicleProjectile(SpawnedProjectile).VehicleDeflectSound, false, true, false);
	}
}

defaultproperties
{
	MIMaxShootingRange=50000; //1km
	OriginalRO2Projectile=class'`{ALBULLETNAME}'

	WeaponProjectiles[0]=class'AL_`{ALBULLETNAME}';
	WeaponProjectiles[1]=class'AL_`{ALBULLETNAME}';
`ifdef(ALHASTRACER)
	TracerClass=class'AL_`{ALBULLETNAME}Tracer'; 
`endif
`ifdef(ALHASNEWATTACH)
	AttachmentClass=class'AL_Attach_`{ALHASNEWATTACH}';
`endif

/**
 *	This is the unique instance of the AnimTree used by this SkeletalMeshComponent.
 *	THIS SHOULD NEVER POINT TO AN ANIMTREE IN A CONTENT PACKAGE.
 *
 * var()	const editinline export	AnimNode	Animations <MaxPropertyDepth=3>;	Begin Object Name=SkeletalMeshComponent0
 **/
`ifdef(ALANIMTREEFIX)
	Begin Object Name=FirstPersonMesh
		Animations=none		//So yeah, it turns out that Epic wasn't kidding about that ^^^
		AnimTreeTemplate=`{ALANIMTREEFIX}
	End Object
`endif

`ifdef(ALHMG)  //hack to turn off HMG firing shake until I can find out why it breaks for me
	FireCameraAnim[0]=none
	FireCameraAnim[1]=none
`endif
} 
