/*	
 *	Antilag is the core mutator itself.  It hooks into ModifyPlayer to replace a newly 
 *	spawned player's weapons to their Antilag counterparts.  It also handles the
 *	assigning and cleanup of AntilagRelays & last death time data.
 *
 *		-Mekhazzio 
 */
 
class Antilag extends Mutator
	config(Antilag);

var config bool  bAdvancedBallistics;
var config float AnnounceMsgTime;

struct AntilagInfo
{
	var Controller 		Controller;	
	var AntilagRelay 	ALRelay;
	var float 		LastDiedTime;
};

var array<AntilagInfo>	AntilagInfoList;

//when the player spawns, replace their weapons with new child classes
function ModifyPlayer(Pawn Other)
{
	local ROPawn 			ROP;
	local ROPlayerController	ROPC;
	local ROWeapon			CurrentWeapon;
	local ROInventoryManager 	InvMgr;
	local array<ROWeapon> 		WeaponList;
	local bool			bFoundAPrimary;
	local string 			ALclassname;
	local int			i;
	local AntilagAnnouncer 		ALAnnouncer;

	ROP = ROPawn(Other);
	if (ROP != none)
		InvMgr = ROInventoryManager(ROP.InvManager);
		
	if (InvMgr != none)
	{
		InvMgr.GetWeaponList(WeaponList);

		bFoundAPrimary = false;
		for(i=0; i < WeaponList.Length; i++)
		{
			CurrentWeapon = WeaponList[i];
			
			if (!bFoundAPrimary && CurrentWeapon.Category != ROIC_Primary)
				continue;  //always do primary weapons first, to make sure the player spawns with primary selected
			
			ALclassname = GetAntilagClassStringForObj(CurrentWeapon);

			if (ALclassname == "") //no AL class found
			{	//workaround for inventory replication issues - just rebuild the item list from scratch
				//ALclassname = "ROGameContent." $ string(CurrentWeapon.class);
				continue;
			}
			//`ALlog(ROP @ "Replacing" @ currentweapon.category @ ALclassname)
			
			CurrentWeapon.DetachWeapon();	
			InvMgr.RemoveFromInventory(CurrentWeapon); //remove the old one  
 			CurrentWeapon.Destroy(); 	
			CurrentWeapon = ROWeapon(ROP.LoadAndCreateInventory(ALclassname)); //add the new one

			if (!bFoundAPrimary)
			{
				bFoundAPrimary = true;
				WeaponList.Remove(i, 1); //pull this out of the weapon list so we don't do it again
				i = -1; //then start the list over	
			}
		}		
		
		ROPC = ROPlayerController(ROP.Controller);
		if (ROPC != none)
		{
			GetMyRelay(ROPC); //make sure the relay is spawned
			if (AnnounceMsgTime > 0 && (WorldInfo.TimeSeconds - ROPC.CreationTime) < AnnounceMsgTime )
			{
				ALAnnouncer = Spawn(class'AntilagAnnouncer');
				if (ALAnnouncer != none) {
					ALAnnouncer.ROPC = ROPC;
					ALAnnouncer.bAdvancedBallistics = bAdvancedBallistics;
				}
			}
		}
	}
	
	Super.ModifyPlayer(Other);
}

//replaces the mounted weapons
function bool IsRelevant(Actor Other)
{
	if(ROTurret_Maxim_HMG(Other) != none)
		ROTurret_Maxim_HMG(Other).DefaultInventoryContentClass[0]="Antilag.AL_Maxim_HMG_Turret_MG";
	else if(ROTurret_MG34_LMG(Other) != none)
		ROTurret_MG34_LMG(Other).DefaultInventoryContentClass[0]="Antilag.AL_MG34_LMG_MG";
		
	return Super.IsRelevant(Other);
}

//overridden to catch player death times, to give humans a short grace period after death to still
//be able to land hits without being post-death invalidated.  Keeps midair bullets from fizzling
function bool PreventDeath(Pawn Killed, Controller Killer, class<DamageType> damageType, vector HitLocation)
{
//	if (PlayerController(Killed.Controller) != none)
	if (Killed.Controller != none)
		UpdateLastDiedTime(Killed.Controller);
	
	return Super.PreventDeath(Killed, Killer, damageType, HitLocation);
}

//spawns a relay and links it into the list
function int AddNewInfo(Controller MyController)
{
	local AntilagInfo NewInfo;
	local AntilagRelay NewRelay;
	
	if (MyController == none)
		return -1;
		
	NewInfo.Controller = MyController;
	NewInfo.LastDiedTime = 0;

	NewRelay = MyController.Spawn(class'AntilagRelay', MyController);
	if (NewRelay == none)
	{
		`ALLog("CRITICAL: Relay spawning failed for" @ MyController)
		return -1;
	}
	
	NewRelay.bAdvancedBallistics = bAdvancedBallistics;
	NewRelay.Controller = MyController;
	if (PlayerController(MyController) != none)
		NewRelay.ClientSetOwner(PlayerController(MyController));

	NewInfo.ALRelay = NewRelay;
	
	AntilagInfoList.AddItem(NewInfo);
	return (AntilagInfoList.Length - 1);
}

//mutators have to do everything in such clunky ways
function UpdateLastDiedTime(Controller MyController)
{
	local int i;

	if (MyController == none)
		return;
		
	i = AntilagInfoList.Find('Controller', MyController);

	if (i < 0)
		i = AddNewInfo(MyController);

	if (i < 0)
		return;
		
	AntilagInfoList[i].LastDiedTime = WorldInfo.TimeSeconds;
}

function float GetLastDiedTime(Controller MyController)
{
	local int i;

	if (MyController == none)
		return 0;
		
	i = AntilagInfoList.Find('Controller', MyController);

	if (i >= 0)
		return AntilagInfoList[i].LastDiedTime;
	else
		return 0;
}

function AntilagRelay GetMyRelay(Controller MyController)
{
	local int i;

	if (MyController == none)
		return none;
		
	i = AntilagInfoList.Find('Controller', MyController);

	if (i < 0)
		i = AddNewInfo(MyController);
		
	return AntilagInfoList[i].ALRelay;
}

//removes any disconnected players from the list, called via looping timer
function DoListCleanup() 
{
	local int i;
	for (i=0; i < AntilagInfoList.Length; i++)
	{
		if ( AntilagInfoList[i].Controller == none)
		{
			AntilagInfoList.Remove(i, 1); //garbage collection will get the relay
			i--;
		}
	}	
}

//returns corresponding AL class string for dynamic load
function string GetAntilagClassStringForObj(ROWeapon startweapon)
{
	local string classname;
	local string levelstr;
	local array<string> strs;
	
	classname = string(startweapon.name);
	ParseStringIntoArray(classname, strs, "_", false);
	levelstr = "_" $ strs[strs.Length-2];
	strs.Remove(strs.Length - 2, 2); //trim the level & trailing object number
	strs.Remove(0, 1); //trim the leading ROWeap, to speed up the later string compares
	JoinArray(strs, classname, "_", false);	
	classname = locs(classname);
	
	//this is probably faster than using a dynamicloadobject test to validate the new object
	switch (classname)
	{
		case "avt40_rifle"		:	return "Antilag.AL_AVT40_Rifle" 	$ levelstr; 
		case "c96_pistol"		:	return "Antilag.AL_C96_Pistol" 		$ levelstr; 
		case "dp28_lmg"			:	return "Antilag.AL_DP28_LMG" 		$ levelstr; 
		case "g41scoped_rifle"		:	return "Antilag.AL_G41Scoped_Rifle" 	$ levelstr; 
		case "g41_rifle"		:	return "Antilag.AL_G41_Rifle" 		$ levelstr; 
		case "kar98scoped_rifle"	:	return "Antilag.AL_Kar98Scoped_Rifle" 	$ levelstr; 
		case "kar98_rifle"		:	return "Antilag.AL_Kar98_Rifle" 	$ levelstr; 
		case "m1895_revolver"		:	return "Antilag.AL_M1895_Revolver" 	$ levelstr; 
		case "mg34_lmg"			:	return "Antilag.AL_MG34_LMG" 		$ levelstr; 
		case "mkb42_assaultrifle"	:	return "Antilag.AL_MKb42_AssaultRifle" 	$ levelstr; 
		case "mn9130scoped_rifle"	:	return "Antilag.AL_MN9130Scoped_Rifle" 	$ levelstr; 
		case "mn9130_rifle"		:	return "Antilag.AL_MN9130_Rifle" 	$ levelstr; 
		case "mp40_smg"			:	return "Antilag.AL_MP40_SMG" 		$ levelstr; 
		case "p38_pistol"		:	return "Antilag.AL_P38_Pistol" 		$ levelstr; 
		case "ppsh41_smg"		:	return "Antilag.AL_PPSH41_SMG" 		$ levelstr; 
		case "ptrs1941_atrifle"		:	return "Antilag.AL_PTRS1941_ATRifle" 	$ levelstr; 
		case "pzb784r_atrifle"		:	return "Antilag.AL_PzB784r_ATRifle" 	$ levelstr; 
		case "svt40scoped_rifle"	:	return "Antilag.AL_SVT40Scoped_Rifle" 	$ levelstr; 
		case "svt40_rifle"		:	return "Antilag.AL_SVT40_Rifle" 	$ levelstr; 
		case "tt33_pistol"		:	return "Antilag.AL_TT33_Pistol" 	$ levelstr; 
	}
	return "";
}

event PostBeginPlay()
{
	SetTimer( `ALCLEANUPTIME , true, 'DoListCleanup');
	`ALlog("Antilag v2 by Mekhazzio is active.")
}

defaultproperties
{
	GroupNames(0)="ANTILAG"
}
